@@ -51,6 +51,7 @@
  * when USE_THREAD_DUMP is set.
  */
 volatile unsigned long threads_to_dump = 0;
+unsigned int panic_started = 0;
 unsigned int debug_commands_issued = 0;
 
 /* dumps a backtrace of the current thread that is appended to buffer <buf>.
@@ -220,6 +221,7 @@ void ha_task_dump(struct buffer *buf, const struct task *task, const char *pfx)
 	const struct stream *s = NULL;
 	const struct appctx __maybe_unused *appctx = NULL;
 	struct hlua __maybe_unused *hlua = NULL;
+	const struct stconn *sc;
 
 	if (!task) {
 		chunk_appendf(buf, "0\n");
@@ -236,8 +238,8 @@ void ha_task_dump(struct buffer *buf, const struct task *task, const char *pfx)
 		              "%p (task) calls=%u last=%llu%s\n",
 		              task,
 		              task->calls,
-		              task->call_date ? (unsigned long long)(now_mono_time() - task->call_date) : 0,
-		              task->call_date ? " ns ago" : "");
+		              task->wake_date ? (unsigned long long)(now_mono_time() - task->wake_date) : 0,
+		              task->wake_date ? " ns ago" : "");
 
 	chunk_appendf(buf, "%s  fct=%p(", pfx, task->process);
 	resolve_sym_name(buf, NULL, task->process);
@@ -250,8 +252,8 @@ void ha_task_dump(struct buffer *buf, const struct task *task, const char *pfx)
 
 	if (task->process == process_stream && task->context)
 		s = (struct stream *)task->context;
-	else if (task->process == task_run_applet && task->context)
-		s = sc_strm(appctx_sc((struct appctx *)task->context));
+	else if (task->process == task_run_applet && task->context && (sc = appctx_sc((struct appctx *)task->context)))
+		s = sc_strm(sc);
 	else if (task->process == sc_conn_io_cb && task->context)
 		s = sc_strm(((struct stconn *)task->context));
 
@@ -330,6 +332,14 @@ static int debug_parse_cli_show_libs(char **args, char *payload, struct appctx *
 /* dumps a state of all threads into the trash and on fd #2, then aborts. */
 void ha_panic()
 {
+	if (HA_ATOMIC_FETCH_ADD(&panic_started, 1) != 0) {
+		/* a panic dump is already in progress, let's not disturb it,
+		 * we'll be called via signal DEBUGSIG. By returning we may be
+		 * able to leave a current signal handler (e.g. WDT) so that
+		 * this will ensure more reliable signal delivery.
+		 */
+		return;
+	}
 	chunk_reset(&trash);
 	chunk_appendf(&trash, "Thread %u is about to kill the process.\n", tid + 1);
 	ha_thread_dump_all_to_trash();
