@@ -15,9 +15,9 @@
 #include <haproxy/api.h>
 #include <haproxy/applet.h>
 #include <haproxy/cli.h>
+#include <haproxy/ssl_ckch.h>
 #include <haproxy/dynbuf.h>
 #include <haproxy/cfgparse.h>
-#include <haproxy/connection.h>
 #include <haproxy/global.h>
 #include <haproxy/istbuf.h>
 #include <haproxy/h1_htx.h>
@@ -25,12 +25,16 @@
 #include <haproxy/http_ana-t.h>
 #include <haproxy/http_client.h>
 #include <haproxy/http_htx.h>
+#include <haproxy/http_rules.h>
 #include <haproxy/htx.h>
 #include <haproxy/log.h>
 #include <haproxy/proxy.h>
+#include <haproxy/resolvers.h>
+#include <haproxy/sc_strm.h>
 #include <haproxy/server.h>
 #include <haproxy/ssl_sock-t.h>
-#include <haproxy/stream_interface.h>
+#include <haproxy/sock_inet.h>
+#include <haproxy/stconn.h>
 #include <haproxy/tools.h>
 
 #include <string.h>
@@ -38,11 +42,21 @@
 
 static struct proxy *httpclient_proxy;
 static struct server *httpclient_srv_raw;
+
 #ifdef USE_OPENSSL
+/* if the httpclient is not configured, error are ignored and features are limited */
+static int hard_error_ssl = 0;
 static struct server *httpclient_srv_ssl;
+static int httpclient_ssl_verify = SSL_SOCK_VERIFY_REQUIRED;
+static char *httpclient_ssl_ca_file = NULL;
 #endif
 static struct applet httpclient_applet;
 
+/* if the httpclient is not configured, error are ignored and features are limited */
+static int hard_error_resolvers = 0;
+static char *resolvers_id = NULL;
+static char *resolvers_prefer = NULL;
+
 /* --- This part of the file implement an HTTP client over the CLI ---
  * The functions will be  starting by "hc_cli" for "httpclient cli"
  */
@@ -53,51 +67,64 @@ static struct applet httpclient_applet;
 #define	HC_CLI_F_RES_BODY       0x04
 #define HC_CLI_F_RES_END        0x08
 
+/* the CLI context for the httpclient command */
+struct hcli_svc_ctx {
+	struct httpclient *hc;  /* the httpclient instance */
+	uint flags;             /* flags from HC_CLI_F_* above */
+};
 
 /* These are the callback used by the HTTP Client when it needs to notify new
- * data, we only sets a flag in the IO handler  */
-
+ * data, we only sets a flag in the IO handler via the svcctx.
+ */
 void hc_cli_res_stline_cb(struct httpclient *hc)
 {
 	struct appctx *appctx = hc->caller;
+	struct hcli_svc_ctx *ctx;
 
 	if (!appctx)
 		return;
 
-	appctx->ctx.cli.i0 |= HC_CLI_F_RES_STLINE;
+	ctx = appctx->svcctx;
+	ctx->flags |= HC_CLI_F_RES_STLINE;
 	appctx_wakeup(appctx);
 }
 
 void hc_cli_res_headers_cb(struct httpclient *hc)
 {
 	struct appctx *appctx = hc->caller;
+	struct hcli_svc_ctx *ctx;
 
 	if (!appctx)
 		return;
 
-	appctx->ctx.cli.i0 |= HC_CLI_F_RES_HDR;
+	ctx = appctx->svcctx;
+	ctx->flags |= HC_CLI_F_RES_HDR;
 	appctx_wakeup(appctx);
 }
 
 void hc_cli_res_body_cb(struct httpclient *hc)
 {
 	struct appctx *appctx = hc->caller;
+	struct hcli_svc_ctx *ctx;
 
 	if (!appctx)
 		return;
 
-	appctx->ctx.cli.i0 |= HC_CLI_F_RES_BODY;
+	ctx = appctx->svcctx;
+	ctx->flags |= HC_CLI_F_RES_BODY;
 	appctx_wakeup(appctx);
 }
 
 void hc_cli_res_end_cb(struct httpclient *hc)
 {
 	struct appctx *appctx = hc->caller;
+	struct hcli_svc_ctx *ctx;
 
 	if (!appctx)
 		return;
 
-	appctx->ctx.cli.i0 |= HC_CLI_F_RES_END;
+	ctx = appctx->svcctx;
+	ctx->flags |= HC_CLI_F_RES_END;
 	appctx_wakeup(appctx);
 }
 
@@ -107,6 +134,7 @@ void hc_cli_res_end_cb(struct httpclient *hc)
  */
 static int hc_cli_parse(char **args, char *payload, struct appctx *appctx, void *private)
 {
+	struct hcli_svc_ctx *ctx = applet_reserve_svcctx(appctx, sizeof(*ctx));
 	struct httpclient *hc;
 	char *err = NULL;
 	enum http_meth_t meth;
@@ -141,8 +169,8 @@ static int hc_cli_parse(char **args, char *payload, struct appctx *appctx, void
 	hc->ops.res_payload = hc_cli_res_body_cb;
 	hc->ops.res_end = hc_cli_res_end_cb;
 
-	appctx->ctx.cli.p0 = hc; /* store the httpclient ptr in the applet */
-	appctx->ctx.cli.i0 = 0;
+	ctx->hc = hc; /* store the httpclient ptr in the applet */
+	ctx->flags = 0;
 
 	if (httpclient_req_gen(hc, hc->req.url, hc->req.meth, NULL, body) != ERR_NONE)
 		goto err;
@@ -166,68 +194,66 @@ err:
  */
 static int hc_cli_io_handler(struct appctx *appctx)
 {
-	struct stream_interface *si = appctx->owner;
-	struct buffer *trash = alloc_trash_chunk();
-	struct httpclient *hc = appctx->ctx.cli.p0;
+	struct hcli_svc_ctx *ctx = appctx->svcctx;
+	struct stconn *sc = appctx_sc(appctx);
+	struct httpclient *hc = ctx->hc;
 	struct http_hdr *hdrs, *hdr;
 
-	if (!trash)
-		goto out;
-	if (appctx->ctx.cli.i0 & HC_CLI_F_RES_STLINE) {
-		chunk_appendf(trash, "%.*s %d %.*s\n", (unsigned int)istlen(hc->res.vsn), istptr(hc->res.vsn),
-			      hc->res.status, (unsigned int)istlen(hc->res.reason), istptr(hc->res.reason));
-		if (ci_putchk(si_ic(si), trash) == -1)
-			si_rx_room_blk(si);
-		appctx->ctx.cli.i0 &= ~HC_CLI_F_RES_STLINE;
-		goto out;
+	if (ctx->flags & HC_CLI_F_RES_STLINE) {
+		chunk_printf(&trash, "%.*s %d %.*s\n", (unsigned int)istlen(hc->res.vsn), istptr(hc->res.vsn),
+			     hc->res.status, (unsigned int)istlen(hc->res.reason), istptr(hc->res.reason));
+		if (applet_putchk(appctx, &trash) == -1)
+			goto more;
+		ctx->flags &= ~HC_CLI_F_RES_STLINE;
 	}
 
-	if (appctx->ctx.cli.i0 & HC_CLI_F_RES_HDR) {
+	if (ctx->flags & HC_CLI_F_RES_HDR) {
+		chunk_reset(&trash);
 		hdrs = hc->res.hdrs;
 		for (hdr = hdrs; isttest(hdr->v); hdr++) {
-			if (!h1_format_htx_hdr(hdr->n, hdr->v, trash))
-				goto out;
+			if (!h1_format_htx_hdr(hdr->n, hdr->v, &trash))
+				goto too_many_hdrs;
 		}
-		if (!chunk_memcat(trash, "\r\n", 2))
-			goto out;
-		if (ci_putchk(si_ic(si), trash) == -1)
-			si_rx_room_blk(si);
-		appctx->ctx.cli.i0 &= ~HC_CLI_F_RES_HDR;
-		goto out;
+		if (!chunk_memcat(&trash, "\r\n", 2))
+			goto too_many_hdrs;
+		if (applet_putchk(appctx, &trash) == -1)
+			goto more;
+		ctx->flags &= ~HC_CLI_F_RES_HDR;
 	}
 
-	if (appctx->ctx.cli.i0 & HC_CLI_F_RES_BODY) {
+	if (ctx->flags & HC_CLI_F_RES_BODY) {
 		int ret;
 
-		ret = httpclient_res_xfer(hc, &si_ic(si)->buf);
-		channel_add_input(si_ic(si), ret); /* forward what we put in the buffer channel */
+		ret = httpclient_res_xfer(hc, sc_ib(sc));
+		channel_add_input(sc_ic(sc), ret); /* forward what we put in the buffer channel */
 
-		if (!httpclient_data(hc)) {/* remove the flag if the buffer was emptied */
-			appctx->ctx.cli.i0 &= ~HC_CLI_F_RES_BODY;
-		}
-		goto out;
+		/* remove the flag if the buffer was emptied */
+		if (httpclient_data(hc))
+			goto more;
+		ctx->flags &= ~HC_CLI_F_RES_BODY;
 	}
 
 	/* we must close only if F_END is the last flag */
-	if (appctx->ctx.cli.i0 ==  HC_CLI_F_RES_END) {
-		si_shutw(si);
-		si_shutr(si);
-		appctx->ctx.cli.i0 &= ~HC_CLI_F_RES_END;
-		goto out;
+	if (ctx->flags ==  HC_CLI_F_RES_END) {
+		ctx->flags &= ~HC_CLI_F_RES_END;
+		goto end;
 	}
 
-out:
-	/* we didn't clear every flags, we should come back to finish things */
-	if (appctx->ctx.cli.i0)
-		si_rx_room_blk(si);
-
-	free_trash_chunk(trash);
+more:
+	if (!ctx->flags)
+		applet_have_no_more_data(appctx);
 	return 0;
+end:
+	return 1;
+
+too_many_hdrs:
+	return cli_err(appctx, "Too many headers.\n");
 }
 
 static void hc_cli_release(struct appctx *appctx)
 {
-	struct httpclient *hc = appctx->ctx.cli.p0;
+	struct hcli_svc_ctx *ctx = appctx->svcctx;
+	struct httpclient *hc = ctx->hc;
 
 	/* Everything possible was printed on the CLI, we can destroy the client */
 	httpclient_stop_and_destroy(hc);
@@ -406,103 +432,122 @@ error:
 	return ret;
 }
 
+/* Set the 'timeout server' in ms for the next httpclient request */
+void httpclient_set_timeout(struct httpclient *hc, int timeout)
+{
+	hc->timeout_server = timeout;
+}
+
+/*
+ * Sets a destination for the httpclient from an HAProxy addr format
+ * This will prevent to determine the destination from the URL
+ * Return 0 in case of success or -1 otherwise.
+ */
+int httpclient_set_dst(struct httpclient *hc, const char *dst)
+{
+	struct sockaddr_storage *sk;
+	char *errmsg = NULL;
+
+	sockaddr_free(&hc->dst);
+	/* 'sk' is statically allocated (no need to be freed). */
+	sk = str2sa_range(dst, NULL, NULL, NULL, NULL, NULL,
+	                  &errmsg, NULL, NULL,
+	                  PA_O_PORT_OK | PA_O_STREAM | PA_O_XPRT | PA_O_CONNECT);
+	if (!sk) {
+		ha_alert("httpclient: Failed to parse destination address in %s\n", errmsg);
+		free(errmsg);
+		return -1;
+	}
+
+	if (!sockaddr_alloc(&hc->dst, sk, sizeof(*sk))) {
+		ha_alert("httpclient: Failed to allocate sockaddr in %s:%d.\n", __FUNCTION__, __LINE__);
+		return -1;
+	}
+
+	return 0;
+}
+
+/*
+ * Return a split URL in <scheme>, <host>, <port>
+ */
+static void httpclient_spliturl(struct ist url, enum http_scheme *scheme,
+                                struct ist *host, int *port)
+{
+	enum http_scheme scheme_tmp = SCH_HTTP;
+	int port_tmp = 0;
+	struct ist scheme_ist, authority_ist, host_ist, port_ist;
+	char *p, *end;
+	struct http_uri_parser parser;
+
+	parser = http_uri_parser_init(url);
+	scheme_ist = http_parse_scheme(&parser);
+
+	if (isteqi(scheme_ist, ist("http://"))){
+		scheme_tmp = SCH_HTTP;
+		port_tmp = 80;
+	} else if (isteqi(scheme_ist, ist("https://"))) {
+		scheme_tmp = SCH_HTTPS;
+		port_tmp = 443;
+	}
+
+	authority_ist = http_parse_authority(&parser, 1);
+	p = end = istend(authority_ist);
+
+	/* look for a port at the end of the authority */
+	while (p > istptr(authority_ist) && isdigit((unsigned char)*--p))
+		;
+
+	if (*p == ':') {
+		host_ist = ist2(istptr(authority_ist), p - istptr(authority_ist));
+		port_ist = istnext(ist2(p, end - p));
+		ist2str(trash.area, port_ist);
+		port_tmp = atoi(trash.area);
+	} else {
+		host_ist = authority_ist;
+	}
+
+	if (scheme)
+		*scheme = scheme_tmp;
+	if (host)
+		*host = host_ist;
+	if (port)
+		*port = port_tmp;
+
+}
 
 /*
  * Start the HTTP client
  * Create the appctx, session, stream and wakeup the applet
  *
- * FIXME: It also fill the sockaddr with the IP address, but currently only IP
- * in the URL are supported, it lacks a resolver.
- *
  * Return the <appctx> or NULL if it failed
  */
 struct appctx *httpclient_start(struct httpclient *hc)
 {
 	struct applet *applet = &httpclient_applet;
 	struct appctx *appctx;
-	struct session *sess;
-	struct stream *s;
-	int len;
-	struct split_url out;
 
 	/* if the client was started and not ended, an applet is already
 	 * running, we shouldn't try anything */
 	if (httpclient_started(hc) && !httpclient_ended(hc))
 		return NULL;
 
-	hc->flags = 0;
-
-	/* parse URI and fill sockaddr_storage */
-	/* FIXME: use a resolver */
-	len = url2sa(istptr(hc->req.url), istlen(hc->req.url), &hc->dst, &out);
-	if (len == -1) {
-		ha_alert("httpclient: cannot parse uri '%s'.\n", istptr(hc->req.url));
-		goto out;
-	}
-
 	/* The HTTP client will be created in the same thread as the caller,
 	 * avoiding threading issues */
-	appctx = appctx_new(applet);
+	appctx = appctx_new_here(applet, NULL);
 	if (!appctx)
 		goto out;
+	appctx->svcctx = hc;
+	hc->flags = 0;
 
-	sess = session_new(httpclient_proxy, NULL, &appctx->obj_type);
-	if (!sess) {
-		ha_alert("httpclient: out of memory in %s:%d.\n", __FUNCTION__, __LINE__);
+	if (appctx_init(appctx) == -1) {
+		ha_alert("httpclient: Failed to initialize appctx %s:%d.\n", __FUNCTION__, __LINE__);
 		goto out_free_appctx;
 	}
-	if ((s = stream_new(sess, &appctx->obj_type, &hc->req.buf)) == NULL) {
-		ha_alert("httpclient: Failed to initialize stream %s:%d.\n", __FUNCTION__, __LINE__);
-		goto out_free_appctx;
-	}
-
-	if (!sockaddr_alloc(&s->si[1].dst, &hc->dst, sizeof(hc->dst))) {
-		ha_alert("httpclient: Failed to initialize stream in %s:%d.\n", __FUNCTION__, __LINE__);
-		goto out_free_stream;
-	}
-
-	/* choose the SSL server or not */
-	switch (out.scheme) {
-		case SCH_HTTP:
-			s->target = &httpclient_srv_raw->obj_type;
-			break;
-		case SCH_HTTPS:
-#ifdef USE_OPENSSL
-			s->target = &httpclient_srv_ssl->obj_type;
-#else
-			ha_alert("httpclient: OpenSSL is not available %s:%d.\n", __FUNCTION__, __LINE__);
-			goto out_free_stream;
-#endif
-			break;
-	}
-
-	s->flags |= SF_ASSIGNED|SF_ADDR_SET;
-	s->si[1].flags |= SI_FL_NOLINGER;
-	s->res.flags |= CF_READ_DONTWAIT;
-
-	/* applet is waiting for data */
-	si_cant_get(&s->si[0]);
-	appctx_wakeup(appctx);
-
-	task_wakeup(s->task, TASK_WOKEN_INIT);
-	hc->appctx = appctx;
-	hc->flags |= HTTPCLIENT_FS_STARTED;
-	appctx->ctx.httpclient.ptr = hc;
-
-	/* The request was transferred when the stream was created. So switch
-	 * directly to REQ_BODY or RES_STLINE state
-	 */
-	appctx->st0 = (hc->ops.req_payload ? HTTPCLIENT_S_REQ_BODY : HTTPCLIENT_S_RES_STLINE);
 
 	return appctx;
 
-out_free_stream:
-	LIST_DELETE(&s->list);
-	pool_free(pool_head_stream, s);
-out_free_sess:
-	session_free(sess);
 out_free_appctx:
-	appctx_free(appctx);
+	appctx_free_on_early_error(appctx);
 out:
 
 	return NULL;
@@ -522,8 +567,10 @@ void httpclient_stop_and_destroy(struct httpclient *hc)
 	if (hc->flags & HTTPCLIENT_FS_ENDED || !(hc->flags & HTTPCLIENT_FS_STARTED)) {
 		httpclient_destroy(hc);
 	} else {
-	/* if the client wasn't stopped, ask for a stop and destroy */
+		/* if the client wasn't stopped, ask for a stop and destroy */
 		hc->flags |= (HTTPCLIENT_FA_AUTOKILL | HTTPCLIENT_FA_STOP);
+		/* the calling applet doesn't exist anymore */
+		hc->caller = NULL;
 		if (hc->appctx)
 			appctx_wakeup(hc->appctx);
 	}
@@ -555,7 +602,7 @@ void httpclient_destroy(struct httpclient *hc)
 	}
 	ha_free(&hc->res.hdrs);
 	b_free(&hc->res.buf);
-
+	sockaddr_free(&hc->dst);
 
 	free(hc);
 
@@ -587,9 +634,9 @@ err:
 
 static void httpclient_applet_io_handler(struct appctx *appctx)
 {
-	struct httpclient *hc = appctx->ctx.httpclient.ptr;
-	struct stream_interface *si = appctx->owner;
-	struct stream *s = si_strm(si);
+	struct httpclient *hc = appctx->svcctx;
+	struct stconn *sc = appctx_sc(appctx);
+	struct stream *s = __sc_strm(sc);
 	struct channel *req = &s->req;
 	struct channel *res = &s->res;
 	struct htx_blk *blk = NULL;
@@ -635,31 +682,44 @@ static void httpclient_applet_io_handler(struct appctx *appctx)
 				/* call the payload callback */
 				{
 					if (hc->ops.req_payload) {
+						struct htx *hc_htx;
 
 						/* call the request callback */
 						hc->ops.req_payload(hc);
-						/* check if the request buffer is empty */
 
+						hc_htx = htx_from_buf(&hc->req.buf);
 						htx = htx_from_buf(&req->buf);
-						if (!htx_is_empty(htx))
-							goto more;
-						/* Here htx_to_buf() will set buffer data to 0 because
-						 * the HTX is empty, and allow us to do an xfer.
-						 */
-						htx_to_buf(htx, &req->buf);
 
-						ret = b_xfer(&req->buf, &hc->req.buf, b_data(&hc->req.buf));
-						if (!ret)
+						if (htx_is_empty(hc_htx))
 							goto more;
 
+						if (htx_is_empty(htx)) {
+							size_t data = hc_htx->data;
+
+							/* Here htx_to_buf() will set buffer data to 0 because
+							 * the HTX is empty, and allow us to do an xfer.
+							 */
+							htx_to_buf(hc_htx, &hc->req.buf);
+							htx_to_buf(htx, &req->buf);
+							b_xfer(&req->buf, &hc->req.buf, b_data(&hc->req.buf));
+							channel_add_input(req, data);
+						} else {
+							struct htx_ret ret;
+
+							ret = htx_xfer_blks(htx, hc_htx, htx_used_space(hc_htx), HTX_BLK_UNUSED);
+							channel_add_input(req, ret.ret);
+
+							/* we must copy the EOM if we empty the buffer */
+							if (htx_is_empty(hc_htx)) {
+								htx->flags |= (hc_htx->flags & HTX_FL_EOM);
+							}
+							htx_to_buf(htx, &req->buf);
+							htx_to_buf(hc_htx, &hc->req.buf);
+						}
+
+
 						if (!b_data(&hc->req.buf))
 							b_free(&hc->req.buf);
-
-						htx = htx_from_buf(&req->buf);
-						if (!htx)
-							goto more;
-
-						channel_add_input(req, htx->data);
 					}
 
 					htx = htx_from_buf(&req->buf);
@@ -669,6 +729,7 @@ static void httpclient_applet_io_handler(struct appctx *appctx)
 					/* if the request contains the HTX_FL_EOM, we finished the request part. */
 					if (htx->flags & HTX_FL_EOM) {
 						req->flags |= CF_EOI;
+						se_fl_set(appctx->sedesc, SE_FL_EOI);
 						appctx->st0 = HTTPCLIENT_S_RES_STLINE;
 					}
 
@@ -694,7 +755,7 @@ static void httpclient_applet_io_handler(struct appctx *appctx)
 				hc->res.vsn = istdup(htx_sl_res_vsn(sl));
 				hc->res.reason = istdup(htx_sl_res_reason(sl));
 				sz = htx_get_blksz(blk);
-				co_set_data(res, co_data(res) - sz);
+				c_rew(res, sz);
 				htx_remove_blk(htx, blk);
 				/* caller callback */
 				if (hc->ops.res_stline)
@@ -730,11 +791,8 @@ static void httpclient_applet_io_handler(struct appctx *appctx)
 						uint32_t sz = htx_get_blksz(blk);
 
 						c_rew(res, sz);
-						blk = htx_remove_blk(htx, blk);
 
-						if (type == HTX_BLK_UNUSED)
-							continue;
-						else if (type == HTX_BLK_HDR) {
+						if (type == HTX_BLK_HDR) {
 							hdrs[hdr_num].n = istdup(htx_get_blk_name(htx, blk));
 							hdrs[hdr_num].v = istdup(htx_get_blk_value(htx, blk));
 							hdr_num++;
@@ -743,8 +801,10 @@ static void httpclient_applet_io_handler(struct appctx *appctx)
 							/* create a NULL end of array and leave the loop */
 							hdrs[hdr_num].n = IST_NULL;
 							hdrs[hdr_num].v = IST_NULL;
+							htx_remove_blk(htx, blk);
 							break;
 						}
+						blk = htx_remove_blk(htx, blk);
 					}
 
 					if (hdr_num) {
@@ -819,7 +879,7 @@ static void httpclient_applet_io_handler(struct appctx *appctx)
 						if (hc->ops.res_payload)
 							hc->ops.res_payload(hc);
 
-						/* cannot copy everything, need to processs */
+						/* cannot copy everything, need to process */
 						if (vlen != blksz)
 							goto process_data;
 					} else {
@@ -849,13 +909,13 @@ static void httpclient_applet_io_handler(struct appctx *appctx)
 
 process_data:
 
-	si_rx_chan_rdy(si);
+	sc_will_read(sc);
 
 	return;
 more:
 	/* There was not enough data in the response channel */
 
-	si_rx_room_blk(si);
+	sc_need_room(sc);
 
 	if (appctx->st0 == HTTPCLIENT_S_RES_END)
 		goto end;
@@ -871,14 +931,112 @@ more:
 	return;
 
 end:
-	si_shutw(si);
-	si_shutr(si);
+	sc_shutw(sc);
+	sc_shutr(sc);
 	return;
 }
 
+static int httpclient_applet_init(struct appctx *appctx)
+{
+	struct httpclient *hc = appctx->svcctx;
+	struct stream *s;
+	struct sockaddr_storage *addr = NULL;
+	struct sockaddr_storage ss_url = {};
+	struct sockaddr_storage *ss_dst;
+	enum obj_type *target = NULL;
+	struct ist host = IST_NULL;
+	enum http_scheme scheme;
+	int port;
+	int doresolve = 0;
+
+
+	/* parse the URL and  */
+	httpclient_spliturl(hc->req.url, &scheme, &host, &port);
+
+	if (hc->dst) {
+		/* if httpclient_set_dst() was used, sets the alternative address */
+		ss_dst = hc->dst;
+	} else {
+		/* set the dst using the host, or 0.0.0.0 to resolve */
+		ist2str(trash.area, host);
+		ss_dst = str2ip2(trash.area, &ss_url, 0);
+		if (!ss_dst) { /* couldn't get an IP from that, try to resolve */
+			doresolve = 1;
+			ss_dst = str2ip2("0.0.0.0", &ss_url, 0);
+		}
+		sock_inet_set_port(ss_dst, port);
+	}
+
+	if (!sockaddr_alloc(&addr, ss_dst, sizeof(*ss_dst)))
+		goto out_error;
+
+	/* choose the SSL server or not */
+	switch (scheme) {
+		case SCH_HTTP:
+			target = &httpclient_srv_raw->obj_type;
+			break;
+		case SCH_HTTPS:
+#ifdef USE_OPENSSL
+			if (httpclient_srv_ssl) {
+				target = &httpclient_srv_ssl->obj_type;
+			} else {
+				ha_alert("httpclient: SSL was disabled (wrong verify/ca-file)!\n");
+				goto out_free_addr;
+			}
+#else
+			ha_alert("httpclient: OpenSSL is not available %s:%d.\n", __FUNCTION__, __LINE__);
+			goto out_free_addr;
+#endif
+			break;
+	}
+
+	if (appctx_finalize_startup(appctx, httpclient_proxy, &hc->req.buf) == -1) {
+		ha_alert("httpclient: Failed to initialize appctx %s:%d.\n", __FUNCTION__, __LINE__);
+		goto out_free_addr;
+	}
+
+	s = appctx_strm(appctx);
+	s->target = target;
+	/* set the "timeout server" */
+	s->req.wto = hc->timeout_server;
+	s->res.rto = hc->timeout_server;
+
+	if (doresolve) {
+		/* in order to do the set-dst we need to put the address on the front */
+		s->scf->dst = addr;
+	} else {
+		/* in cases we don't use the resolve we already have the address
+		 * and must put it on the backend side, some of the cases are
+		 * not meant to be used on the frontend (sockpair, unix socket etc.) */
+		s->scb->dst = addr;
+	}
+
+	s->scb->flags |= SC_FL_NOLINGER;
+	s->flags |= SF_ASSIGNED;
+	s->res.flags |= CF_READ_DONTWAIT;
+
+	/* applet is waiting for data */
+	applet_need_more_data(appctx);
+	appctx_wakeup(appctx);
+
+	hc->appctx = appctx;
+	hc->flags |= HTTPCLIENT_FS_STARTED;
+
+	/* The request was transferred when the stream was created. So switch
+	 * directly to REQ_BODY or RES_STLINE state
+	 */
+	appctx->st0 = (hc->ops.req_payload ? HTTPCLIENT_S_REQ_BODY : HTTPCLIENT_S_RES_STLINE);
+	return 0;
+
+ out_free_addr:
+	sockaddr_free(&addr);
+ out_error:
+	return -1;
+}
+
 static void httpclient_applet_release(struct appctx *appctx)
 {
-	struct httpclient *hc = appctx->ctx.httpclient.ptr;
+	struct httpclient *hc = appctx->svcctx;
 
 	/* mark the httpclient as ended */
 	hc->flags |= HTTPCLIENT_FS_ENDED;
@@ -902,31 +1060,85 @@ static struct applet httpclient_applet = {
 	.obj_type = OBJ_TYPE_APPLET,
 	.name = "<HTTPCLIENT>",
 	.fct = httpclient_applet_io_handler,
+	.init = httpclient_applet_init,
 	.release = httpclient_applet_release,
 };
 
+
+static int httpclient_resolve_init()
+{
+	struct act_rule *rule;
+	int i;
+	char *do_resolve = NULL;
+	char *http_rules[][11] = {
+	       { "set-var(txn.hc_ip)", "dst", "" },
+	       { do_resolve, "hdr(Host),host_only", "if", "{", "var(txn.hc_ip)", "-m", "ip", "0.0.0.0", "}", "" },
+	       { "return", "status", "503", "if", "{", "var(txn.hc_ip)", "-m", "ip", "0.0.0.0", "}", "" },
+	       { "capture", "var(txn.hc_ip)", "len", "40", "" },
+	       { "set-dst", "var(txn.hc_ip)", "" },
+	       { "" }
+	};
+
+	if (!resolvers_id)
+		resolvers_id = strdup("default");
+
+	memprintf(&do_resolve, "do-resolve(txn.hc_ip,%s%s%s)", resolvers_id, resolvers_prefer ? "," : "", resolvers_prefer ? resolvers_prefer : "");
+	http_rules[1][0] = do_resolve;
+
+	/* Try to create the default resolvers section */
+	resolvers_create_default();
+
+	/* if the resolver does not exist and no hard_error was set, simply ignore resolving */
+	if (!find_resolvers_by_id(resolvers_id) && !hard_error_resolvers) {
+		free(do_resolve);
+		return 0;
+	}
+
+
+	for (i = 0; *http_rules[i][0] != '\0'; i++) {
+		rule = parse_http_req_cond((const char **)http_rules[i], "httpclient", 0, httpclient_proxy);
+		if (!rule) {
+			free(do_resolve);
+			ha_alert("Couldn't setup the httpclient resolver.\n");
+			return 1;
+		}
+		LIST_APPEND(&httpclient_proxy->http_req_rules, &rule->list);
+	}
+
+	free(do_resolve);
+	return 0;
+}
+
+
+
 /*
  * Initialize the proxy for the HTTP client with 2 servers, one for raw HTTP,
  * the other for HTTPS.
  */
-
-static int httpclient_init()
+static int httpclient_precheck()
 {
-	int err_code = 0;
+	int err_code = ERR_NONE;
 	char *errmsg = NULL;
 
+	if (global.mode & MODE_MWORKER_WAIT)
+		return ERR_NONE;
+
 	httpclient_proxy = alloc_new_proxy("<HTTPCLIENT>", PR_CAP_LISTEN|PR_CAP_INT, &errmsg);
 	if (!httpclient_proxy) {
+		memprintf(&errmsg, "couldn't allocate proxy.");
 		err_code |= ERR_ALERT | ERR_FATAL;
 		goto err;
 	}
 
 	proxy_preset_defaults(httpclient_proxy);
 
+	httpclient_proxy->options |= PR_O_WREQ_BODY;
+	httpclient_proxy->retry_type |= PR_RE_CONN_FAILED | PR_RE_DISCONNECTED | PR_RE_TIMEOUT;
 	httpclient_proxy->options2 |= PR_O2_INDEPSTR;
 	httpclient_proxy->mode = PR_MODE_HTTP;
 	httpclient_proxy->maxconn = 0;
 	httpclient_proxy->accept = NULL;
+	httpclient_proxy->conn_retries = CONN_RETRIES;
 	httpclient_proxy->timeout.client = TICK_ETERNITY;
 	/* The HTTP Client use the "option httplog" with the global log server */
 	httpclient_proxy->conf.logformat_string = default_http_log_format;
@@ -935,17 +1147,22 @@ static int httpclient_init()
 	/* clear HTTP server */
 	httpclient_srv_raw = new_server(httpclient_proxy);
 	if (!httpclient_srv_raw) {
-		err_code |= ERR_ALERT | ERR_FATAL;
 		memprintf(&errmsg, "out of memory.");
+		err_code |= ERR_ALERT | ERR_FATAL;
 		goto err;
 	}
 
+	srv_settings_cpy(httpclient_srv_raw, &httpclient_proxy->defsrv, 0);
 	httpclient_srv_raw->iweight = 0;
 	httpclient_srv_raw->uweight = 0;
 	httpclient_srv_raw->xprt = xprt_get(XPRT_RAW);
+	httpclient_srv_raw->flags |= SRV_F_MAPPORTS;  /* needed to apply the port change with resolving */
 	httpclient_srv_raw->id = strdup("<HTTPCLIENT>");
-	if (!httpclient_srv_raw->id)
+	if (!httpclient_srv_raw->id) {
+		memprintf(&errmsg, "out of memory.");
+		err_code |= ERR_ALERT | ERR_FATAL;
 		goto err;
+	}
 
 #ifdef USE_OPENSSL
 	/* SSL HTTP server */
@@ -955,70 +1172,113 @@ static int httpclient_init()
 		err_code |= ERR_ALERT | ERR_FATAL;
 		goto err;
 	}
+	srv_settings_cpy(httpclient_srv_ssl, &httpclient_proxy->defsrv, 0);
 	httpclient_srv_ssl->iweight = 0;
 	httpclient_srv_ssl->uweight = 0;
 	httpclient_srv_ssl->xprt = xprt_get(XPRT_SSL);
 	httpclient_srv_ssl->use_ssl = 1;
+	httpclient_srv_ssl->flags |= SRV_F_MAPPORTS;  /* needed to apply the port change with resolving */
 	httpclient_srv_ssl->id = strdup("<HTTPSCLIENT>");
-	if (!httpclient_srv_ssl->id)
+	if (!httpclient_srv_ssl->id) {
+		memprintf(&errmsg, "out of memory.");
+		err_code |= ERR_ALERT | ERR_FATAL;
 		goto err;
+	}
+
+	httpclient_srv_ssl->ssl_ctx.verify = httpclient_ssl_verify;
+	/* if the verify is required, try to load the system CA */
+	if (httpclient_ssl_verify == SSL_SOCK_VERIFY_REQUIRED) {
+
+		if (!httpclient_ssl_ca_file)
+			httpclient_ssl_ca_file = strdup("@system-ca");
+
+		httpclient_srv_ssl->ssl_ctx.ca_file = httpclient_ssl_ca_file;
+		if (!ssl_store_load_locations_file(httpclient_srv_ssl->ssl_ctx.ca_file, 1, CAFILE_CERT)) {
+			/* if we failed to load the ca-file, only quits in
+			 * error with hard_error, otherwise just disable the
+			 * feature. */
+			if (hard_error_ssl) {
+				memprintf(&errmsg, "cannot initialize SSL verify with 'ca-file \"%s\"'.", httpclient_srv_ssl->ssl_ctx.ca_file);
+				err_code |= ERR_ALERT | ERR_FATAL;
+				goto err;
+			} else {
+				ha_free(&httpclient_srv_ssl->ssl_ctx.ca_file);
+				srv_drop(httpclient_srv_ssl);
+				httpclient_srv_ssl = NULL;
+			}
+		}
+	}
 
-	httpclient_srv_ssl->ssl_ctx.verify = SSL_SOCK_VERIFY_NONE;
 #endif
 
 	/* add the proxy in the proxy list only if everything is successful */
 	httpclient_proxy->next = proxies_list;
 	proxies_list = httpclient_proxy;
 
+	if (httpclient_resolve_init() != 0) {
+		memprintf(&errmsg, "cannot initialize resolvers.");
+		err_code |= ERR_ALERT | ERR_FATAL;
+		goto err;
+	}
+
 	/* link the 2 servers in the proxy */
 	httpclient_srv_raw->next = httpclient_proxy->srv;
 	httpclient_proxy->srv = httpclient_srv_raw;
 
 #ifdef USE_OPENSSL
-	httpclient_srv_ssl->next = httpclient_proxy->srv;
-	httpclient_proxy->srv = httpclient_srv_ssl;
+	if (httpclient_srv_ssl) {
+		httpclient_srv_ssl->next = httpclient_proxy->srv;
+		httpclient_proxy->srv = httpclient_srv_ssl;
+	}
 #endif
 
 
-	return 0;
-
 err:
-	ha_alert("httpclient: cannot initialize.\n");
-	free(errmsg);
-	srv_drop(httpclient_srv_raw);
+	if (err_code & ERR_CODE) {
+		ha_alert("httpclient: cannot initialize: %s\n", errmsg);
+		free(errmsg);
+		srv_drop(httpclient_srv_raw);
 #ifdef USE_OPENSSL
-	srv_drop(httpclient_srv_ssl);
+		srv_drop(httpclient_srv_ssl);
 #endif
-	free_proxy(httpclient_proxy);
+		free_proxy(httpclient_proxy);
+	}
 	return err_code;
 }
 
-static int httpclient_cfg_postparser()
+static int httpclient_postcheck()
 {
+	int err_code = ERR_NONE;
 	struct logsrv *logsrv;
 	struct proxy *curproxy = httpclient_proxy;
 	char *errmsg = NULL;
 
+	if (global.mode & MODE_MWORKER_WAIT)
+		return ERR_NONE;
+
 	/* copy logs from "global" log list */
 	list_for_each_entry(logsrv, &global.logsrvs, list) {
 		struct logsrv *node = malloc(sizeof(*node));
 
 		if (!node) {
-			ha_alert("httpclient: cannot allocate memory.\n");
+			memprintf(&errmsg, "out of memory.");
+			err_code |= ERR_ALERT | ERR_FATAL;
 			goto err;
 		}
 
 		memcpy(node, logsrv, sizeof(*node));
 		LIST_INIT(&node->list);
 		LIST_APPEND(&curproxy->logsrvs, &node->list);
+		node->ring_name = logsrv->ring_name ? strdup(logsrv->ring_name) : NULL;
+		node->conf.file = logsrv->conf.file ? strdup(logsrv->conf.file) : NULL;
 	}
 	if (curproxy->conf.logformat_string) {
 		curproxy->conf.args.ctx = ARGC_LOG;
 		if (!parse_logformat_string(curproxy->conf.logformat_string, curproxy, &curproxy->logformat,
 					    LOG_OPT_MANDATORY|LOG_OPT_MERGE_SPACES,
 					    SMP_VAL_FE_LOG_END, &errmsg)) {
-			ha_alert("httpclient: failed to parse log-format : %s.\n", errmsg);
-			free(errmsg);
+			memprintf(&errmsg, "failed to parse log-format : %s.", errmsg);
+			err_code |= ERR_ALERT | ERR_FATAL;
 			goto err;
 		}
 		curproxy->conf.args.file = NULL;
@@ -1026,27 +1286,120 @@ static int httpclient_cfg_postparser()
 	}
 
 #ifdef USE_OPENSSL
-	{
-		int err_code = 0;
-
+	if (httpclient_srv_ssl) {
 		/* init the SNI expression */
 		/* always use the host header as SNI, without the port */
 		httpclient_srv_ssl->sni_expr = strdup("req.hdr(host),field(1,:)");
 		err_code |= server_parse_sni_expr(httpclient_srv_ssl, httpclient_proxy, &errmsg);
 		if (err_code & ERR_CODE) {
-			ha_alert("httpclient: failed to configure sni: %s.\n", errmsg);
-			free(errmsg);
+			memprintf(&errmsg, "failed to configure sni: %s.", errmsg);
 			goto err;
 		}
 	}
 #endif
 
-	return 0;
 err:
-	return 1;
+	if (err_code & ERR_CODE) {
+		ha_alert("httpclient: failed to initialize: %s\n", errmsg);
+		free(errmsg);
+
+	}
+	return err_code;
 }
 
 /* initialize the proxy and servers for the HTTP client */
 
-INITCALL0(STG_REGISTER, httpclient_init);
-REGISTER_CONFIG_POSTPARSER("httpclient", httpclient_cfg_postparser);
+REGISTER_PRE_CHECK(httpclient_precheck);
+REGISTER_POST_CHECK(httpclient_postcheck);
+
+static int httpclient_parse_global_resolvers(char **args, int section_type, struct proxy *curpx,
+                                        const struct proxy *defpx, const char *file, int line,
+                                        char **err)
+{
+	if (too_many_args(1, args, err, NULL))
+		return -1;
+
+	/* any configuration should set the hard_error flag */
+	hard_error_resolvers = 1;
+
+	free(resolvers_id);
+	resolvers_id = strdup(args[1]);
+
+	return 0;
+}
+
+static int httpclient_parse_global_prefer(char **args, int section_type, struct proxy *curpx,
+                                        const struct proxy *defpx, const char *file, int line,
+                                        char **err)
+{
+	if (too_many_args(1, args, err, NULL))
+		return -1;
+
+	/* any configuration should set the hard_error flag */
+	hard_error_resolvers = 1;
+
+
+	if (strcmp(args[1],"ipv4") == 0)
+		resolvers_prefer = "ipv4";
+	else if (strcmp(args[1],"ipv6") == 0)
+		resolvers_prefer = "ipv6";
+	else {
+		ha_alert("parsing [%s:%d] : '%s' expects 'ipv4' or 'ipv6' as argument.\n", file, line, args[0]);
+		return -1;
+	}
+
+	return 0;
+}
+
+
+#ifdef USE_OPENSSL
+static int httpclient_parse_global_ca_file(char **args, int section_type, struct proxy *curpx,
+                                        const struct proxy *defpx, const char *file, int line,
+                                        char **err)
+{
+	if (too_many_args(1, args, err, NULL))
+		return -1;
+
+	/* any configuration should set the hard_error flag */
+	hard_error_ssl = 1;
+
+	free(httpclient_ssl_ca_file);
+	httpclient_ssl_ca_file = strdup(args[1]);
+
+	return 0;
+}
+
+static int httpclient_parse_global_verify(char **args, int section_type, struct proxy *curpx,
+                                        const struct proxy *defpx, const char *file, int line,
+                                        char **err)
+{
+	if (too_many_args(1, args, err, NULL))
+		return -1;
+
+	/* any configuration should set the hard_error flag */
+	hard_error_ssl = 1;
+
+	if (strcmp(args[1],"none") == 0)
+		httpclient_ssl_verify = SSL_SOCK_VERIFY_NONE;
+	else if (strcmp(args[1],"required") == 0)
+		httpclient_ssl_verify = SSL_SOCK_VERIFY_REQUIRED;
+	else {
+		ha_alert("parsing [%s:%d] : '%s' expects 'none' or 'required' as argument.\n", file, line, args[0]);
+		return -1;
+	}
+
+	return 0;
+}
+#endif /* ! USE_OPENSSL */
+
+static struct cfg_kw_list cfg_kws = {ILH, {
+	{ CFG_GLOBAL, "httpclient.resolvers.id", httpclient_parse_global_resolvers },
+	{ CFG_GLOBAL, "httpclient.resolvers.prefer", httpclient_parse_global_prefer },
+#ifdef USE_OPENSSL
+	{ CFG_GLOBAL, "httpclient.ssl.verify", httpclient_parse_global_verify },
+	{ CFG_GLOBAL, "httpclient.ssl.ca-file", httpclient_parse_global_ca_file },
+#endif
+	{ 0, NULL, NULL },
+}};
+
+INITCALL1(STG_REGISTER, cfg_register_keywords, &cfg_kws);
